﻿/**
 * AGreement Analysis Toolkit, AGATe v2.0 (.NET Framework 4.5 C# version)
 * 
 * 	    Radu-Daniel Vatavu, Ph.D.
 * 	    MintViz Lab | MANSiD Research Center
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 * 	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publications for the Agreement Analysis toolkit, and what should be 
 * used to cite it, are:
 *
 *  Radu-Daniel Vatavu, Jacob O. Wobbrock. (2016). 
 *  Between-subjects elicitation studies: Formalization and tool support.
 *  In Proceedings of CHI '16, the 34th ACM Conference on Human Factors in Computing Systems
 *  (San Jose, California, USA, May 2016). New York: ACM Press.
 *	
 *  Radu-Daniel Vatavu, Jacob O. Wobbrock. (2015). 
 *  Formalizing Agreement Analysis for Elicitation Studies: New Measures, Significance Test, and Toolkit.
 *  In Proceedings of CHI'15, the 33rd ACM SIGCHI Conference on Human Factors in Computing Systems 
 *  (Seoul, Republic of Korea, April 2015). New York: ACM Press.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2015-2016, Radu-Daniel Vatavu and Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, nor 
 *	    University of Washington, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Jacob O. Wobbrock 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/

using System;
using System.Collections.Generic;
using System.IO;
using System.Windows.Forms;
using AgreementRates;

namespace AGATe
{
    public partial class FormMain : Form
    {
        private List<string> referentNames = null;              // stores the names of all referents
        private List<string[]> proposalsPerReferent = null;     // participants' proposals stored as a string[] for each referent in the List<>
        
        #region form constructor

        public FormMain()
        {
            InitializeComponent();
            tbLog.ScrollBars = ScrollBars.Both;
            closeToolStripMenuItem.Enabled = false;
            statisticsToolStripMenuItem.Enabled = false;
        }

        #endregion

        #region form menu and actions: open file, close file, exit, clear log

        /// <summary>
        /// Menu option: load elicitation data from file.
        /// </summary>
        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            openFileDialog1.InitialDirectory = Application.StartupPath;
            if (openFileDialog1.ShowDialog() == System.Windows.Forms.DialogResult.OK)
                LoadData(openFileDialog1.FileName);
        }

        /// <summary>
        /// Load elicitation data from file.
        /// </summary>
        private void LoadData(string fileName)
        {
            listView.Items.Clear();
            listView.Columns.Clear();
            this.proposalsPerReferent = new List<string[]>();
            this.referentNames = new List<string>();
            using (StreamReader reader = new StreamReader(fileName))
            {
                // read column names
                string line = reader.ReadLine();
                string[] columnNames = line.Split(',');
                for (int i = 0; i < columnNames.Length; i++)
                    listView.Columns.Add(columnNames[i]);

                // read participants' proposals for each referent
                while ((line = reader.ReadLine()) != null)
                {
                    if (line.Trim().Length == 0) continue;

                    string[] values = line.Split(',');
                    for (int i = 0; i < values.Length; i++)
                        values[i] = values[i].Trim();
                    listView.Items.Add(new ListViewItem(values));

                    referentNames.Add(values[0]);
                    string[] proposals = new string[values.Length - 1];
                    Array.Copy(values, 1, proposals, 0, proposals.Length);
                    proposalsPerReferent.Add(proposals);
                }
            }
            listView.AutoResizeColumns(ColumnHeaderAutoResizeStyle.HeaderSize);
            closeToolStripMenuItem.Enabled = true;
            statisticsToolStripMenuItem.Enabled = true;
        }

        /// <summary>
        /// Menu option: close file.
        /// </summary>
        private void closeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            listView.Items.Clear();
            listView.Columns.Clear();
            proposalsPerReferent = new List<string[]>();
            referentNames = new List<string>();
            statisticsToolStripMenuItem.Enabled = false;
            closeToolStripMenuItem.Enabled = false;
        }

        /// <summary>
        /// Menu option: exit application.
        /// </summary>
        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        /// <summary>
        /// Clear log.
        /// </summary>
        private void btnClearLog_Click(object sender, EventArgs e)
        {
            tbLog.Text = "";
        }

        #endregion

        #region menu options: compute agreement, disagreement, coagreement rates, and significance tests (Vrd and Vbg)

        /// <summary>
        /// Menu option: compute Agreement Rates
        /// </summary>
        private void agreementRateToolStripMenuItem_Click(object sender, EventArgs e)
        {
            tbLog.Text += "-------------------------------\r\n"; 
            tbLog.Text += ">> Computing agreement rates...\r\n";
            tbLog.Text += "-------------------------------\r\n";
            double avgAR = 0;
            for (int i = 0; i < referentNames.Count; i++)
            {
                double[] CI95 = null;
                double ar = AgreementMeasures.AR(proposalsPerReferent[i], out CI95, 0.05);
                tbLog.Text += String.Format("AR({0}) = {1:.000}, CI 95% = [{2:.000},{3:.000}]\r\n", 
                    referentNames[i], ar,
                    CI95[0], CI95[1]
                );
                avgAR += ar;
            }
            avgAR /= referentNames.Count;
            tbLog.Text += "----------------------------\r\n";
            tbLog.Text += String.Format("average AR = {0:.000}\r\n", avgAR);
            tbLog.Text += "----------------------------\r\n\r\n";
            tbLog.SelectionStart = tbLog.Text.Length;
            tbLog.ScrollToCaret();
        }

        /// <summary>
        /// Menu option: compute Disagreement Rates
        /// </summary>
        private void disagreementRateToolStripMenuItem_Click(object sender, EventArgs e)
        {
            tbLog.Text += "----------------------------------\r\n";
            tbLog.Text += ">> Computing disagreement rates...\r\n";
            tbLog.Text += "----------------------------------\r\n";
            double avgDR = 0;
            for (int i = 0; i < referentNames.Count; i++)
            {
                double dr = AgreementMeasures.DR(proposalsPerReferent[i]);
                tbLog.Text += String.Format("DR({0}) = {1:.000}\r\n", referentNames[i], dr);
                avgDR += dr;
            }
            avgDR /= referentNames.Count;
            tbLog.Text += "-------------------------------\r\n";
            tbLog.Text += String.Format("average DR = {0:.000}\r\n", avgDR);
            tbLog.Text += "-------------------------------\r\n\r\n";
            tbLog.SelectionStart = tbLog.Text.Length;
            tbLog.ScrollToCaret();
        }

        /// <summary>
        /// Menu option: compute Coagreement between K referents
        /// </summary>
        private void coagreementRateToolStripMenuItem_Click(object sender, EventArgs e)
        {
            (new CoagreementForm(referentNames, proposalsPerReferent, tbLog)).ShowDialog();
        }

        /// <summary>
        /// Menu option: run statistical significance test (Vrd) against zero
        /// </summary>
        private void sampleagainstZeroToolStripMenuItem_Click(object sender, EventArgs e)
        {
            (new ZeroForm(referentNames, proposalsPerReferent, tbLog)).ShowDialog();
        }

        /// <summary>
        /// Menu option: run statistical significance test (Vrd) for 2 referents
        /// </summary>
        private void pairedTestsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            (new _2RelatedSamplesForm(referentNames, proposalsPerReferent, tbLog)).ShowDialog();
        }

        /// <summary>
        /// Menu option: run statistical significance test (Vrd) for K referents
        /// </summary>
        private void allReferentsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            (new KRelatedSamplesForm(referentNames, proposalsPerReferent, tbLog)).ShowDialog();
        }

        /// <summary>
        /// Menu option: run statistical significance test (Vb) for G groups of participants
        /// </summary>
        private void gIndependentGroupsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            (new GIndependentSamplesForm(referentNames, proposalsPerReferent, tbLog)).ShowDialog();
        }
        
        #endregion
    }
}