﻿/**
 * AgreementRates.DLL (.NET Framework 4.5 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 * 	    MintViz Lab | MANSiD Research Center
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 * 	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publications for the Agreement Rates toolkit, and what should be 
 * used to cite it, are:
 *
 *  Radu-Daniel Vatavu, Jacob O. Wobbrock. (2016). 
 *  Between-subjects elicitation studies: Formalization and tool support.
 *  In Proceedings of CHI '16, the 34th ACM Conference on Human Factors in Computing Systems
 *  (San Jose, California, USA, May 2016). New York: ACM Press.
 *	
 *  Radu-Daniel Vatavu, Jacob O. Wobbrock. (2015). 
 *  Formalizing Agreement Analysis for Elicitation Studies: New Measures, Significance Test, and Toolkit.
 *  In Proceedings of CHI'15, the 33rd ACM SIGCHI Conference on Human Factors in Computing Systems 
 *  (Seoul, Republic of Korea, April 2015). New York: ACM Press.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2015-206, Radu-Daniel Vatavu and Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, nor 
 *	    University of Washington, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Jacob O. Wobbrock 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/

using System;
using System.Collections.Generic;

namespace AgreementRates
{
    /// <summary>
    /// 
    /// Implements functions to compute Agreement, Disagreement, and Coagreement Rates,
    /// and Confidence Intervals for Agreement Rates.
    /// 
    /// </summary>
    public class AgreementMeasures
    {
        private delegate double AgreementMeasure(string[] participantsProposal);

        #region agreement rate

        /// <summary>
        /// Computes the agreement rate AR(r) for a partition p[] of
        /// the set P of all proposals elicited for referent r.
        /// Example: int[] p = { 15, 5 } then AR = 0.60526
        /// Example: int[] p = { 15, 3, 2 }; then AR = 0.57368
        /// </summary>
        public static double AR(int[] frequencies)
        {
            if (frequencies == null || frequencies.Length == 0)
                throw new ArgumentNullException();
            if (frequencies.Length == 1)
                return 1.0;

            // compute the total number of proposals
            double P = 0;
            for (int i = 0; i < frequencies.Length; i++)
                P += frequencies[i];

            // compute the sum of square ratios
            double AR = 0;
            for (int i = 0; i < frequencies.Length; i++)
                AR += (frequencies[i] / P) * (frequencies[i] / P);

            // apply correcting factors
            return P / (P - 1) * AR - 1 / (P - 1);
        }

        /// <summary>
        /// Computes the agreement rate AR(r) from a set of proposals for referent r.
        /// Example: string[] p = { "A", "A", "B", "A", "B", "A", "A", "B", "A", "A", "B", "B", "A", "A", "A", "A", "A", "A", "A", "A" } then AR = 0.60526
        /// Example: string[] p = { "A", "A", "B", "A", "C", "A", "A", "B", "A", "A", "B", "C", "A", "A", "A", "A", "A", "A", "A", "A" } then AR = 0.57368
        /// </summary>
        public static double AR(string[] participantsProposals)
        {
            if (participantsProposals == null || participantsProposals.Length <= 1)
                throw new ArgumentNullException();

            int n = participantsProposals.Length;
            double pairsInAgreement = 0.0;
            for (int i = 0; i < n; i++)
                for (int j = i + 1; j < n; j++)
                    if (participantsProposals[i].CompareTo(participantsProposals[j]) == 0)
                        pairsInAgreement++;
            return pairsInAgreement / (n * (n - 1) / 2.0);
        }

        /// <summary>
        /// Computes the agreement rate AR(r) from a set of proposals for referent r 
        /// together with the confidence interval (1 - alpha).
        /// Example: string[] p = { "A", "A", "B", "A", "B", "A", "A", "B", "A", "A", "B", "B", "A", "A", "A", "A", "A", "A", "A", "A" } then AR = 0.60526
        /// Example: string[] p = { "A", "A", "B", "A", "C", "A", "A", "B", "A", "A", "B", "C", "A", "A", "A", "A", "A", "A", "A", "A" } then AR = 0.57368
        /// </summary>
        public static double AR(string[] participantsProposals, out double[] confidenceInterval, double alpha = 0.05)
        {
            confidenceInterval = ComputeConfidenceInterval(participantsProposals, AgreementMeasures.AR, alpha);
            return AgreementMeasures.AR(participantsProposals);
        }

        #endregion

        #region disagreement rate

        /// <summary>
        /// Computes the disagreement rate DR(r) for a partition p[] of
        /// the set P of all proposals elicited for referent r.
        /// Example: int[] p = { 15, 5 } then DR = 0.39474
        /// Example: int[] p = { 15, 3, 2 }; then DR = 0.42632
        /// </summary>
        public static double DR(int[] p)
        {
            return 1.0 - AR(p);
        }

        /// <summary>
        /// Computes the disagreement rate DR(r) from a set of proposals for referent r.
        /// Example: string[] p = { "A", "A", "B", "A", "B", "A", "A", "B", "A", "A", "B", "B", "A", "A", "A", "A", "A", "A", "A", "A" } then DR = 0.39474
        /// Example: string[] p = { "A", "A", "B", "A", "C", "A", "A", "B", "A", "A", "B", "C", "A", "A", "A", "A", "A", "A", "A", "A" } then DR = 0.42632
        /// </summary>
        public static double DR(string[] participantsProposals)
        {
            return 1.0 - AR(participantsProposals);
        }

        /// <summary>
        /// Computes the disagreement rate DR(r) from a set of proposals for referent r
        /// and the confidence interval (1-alpha).
        /// Example: string[] p = { "A", "A", "B", "A", "B", "A", "A", "B", "A", "A", "B", "B", "A", "A", "A", "A", "A", "A", "A", "A" } then DR = 0.39474
        /// Example: string[] p = { "A", "A", "B", "A", "C", "A", "A", "B", "A", "A", "B", "C", "A", "A", "A", "A", "A", "A", "A", "A" } then DR = 0.42632
        /// </summary>
        public static double DR(string[] participantsProposals, out double[] confidenceInterval, double alpha = 0.05)
        {
            confidenceInterval = ComputeConfidenceInterval(participantsProposals, AgreementMeasures.DR, alpha);
            return AgreementMeasures.DR(participantsProposals);
        }

        #endregion

        #region coagreement rate (within-subjects)

        /// <summary>
        /// Computes the coagreement rate AR(r1,r2) for two referents r1 and r2
        /// for which proposals were elicited from participants.
        /// </summary>
        public static double CR(string[] participantsProposals_r1, string[] participantsProposals_r2)
        {
            if (participantsProposals_r1 == null || participantsProposals_r1.Length <= 1 ||
                participantsProposals_r2 == null || participantsProposals_r2.Length <= 1)
                throw new ArgumentNullException();
            if (participantsProposals_r1.Length != participantsProposals_r2.Length)
                throw new ArgumentException(String.Format("Array lengths do not match: {0} vs. {1}", participantsProposals_r1.Length, participantsProposals_r2.Length));

            int n = participantsProposals_r1.Length;
            double pairsInAgreement = 0.0;
            for (int i = 0; i < n; i++)
                for (int j = i + 1; j < n; j++)
                    if ((participantsProposals_r1[i].CompareTo(participantsProposals_r1[j]) == 0) &&    // participants i and j are in agreement for referent r1
                        (participantsProposals_r2[i].CompareTo(participantsProposals_r2[j]) == 0))      // participants i and j are in agreement for referent r2 as well
                        pairsInAgreement++;
            return pairsInAgreement / (n * (n - 1) / 2.0);
        }

        /// <summary>
        /// Computes the coagreement rate CR(r1,r2,...,rk) for k referents
        /// for which proposals were elicited from participants.
        /// </summary>
        public static double CR(List<string[]> participantsProposals)
        {
            if (participantsProposals == null)
                throw new ArgumentNullException();
            for (int i = 0; i < participantsProposals.Count; i++)
                if (participantsProposals[i] == null)
                    throw new ArgumentNullException();
                else
                    if (participantsProposals[i].Length != participantsProposals[0].Length)
                        throw new ArgumentException(String.Format("Array lengths do not match.", participantsProposals[0].Length, participantsProposals[i].Length));

            int n = participantsProposals[0].Length;
            double pairsInAgreement = 0.0;
            for (int i = 0; i < n; i++)
                for (int j = i + 1; j < n; j++)
                {
                    bool all_K_InAgreement = true;
                    for (int k = 0; k < participantsProposals.Count; k++)
                        if (participantsProposals[k][i].CompareTo(participantsProposals[k][j]) != 0)
                        {
                            all_K_InAgreement = false;
                            break;
                        }
                    if (all_K_InAgreement)
                        pairsInAgreement++;
                }
            return pairsInAgreement / (n * (n - 1) / 2.0);
        }

        #endregion

        #region coagreement rate (between-subjects)

        /// <summary>
        /// Computes the coagreement rate for a given referent for k independent 
        /// groups of participants: CRb(G1,G2,...Gk).
        /// </summary>
        public static double CRb(List<string[]> participantsProposals)
        {
            if (participantsProposals == null || participantsProposals.Count < 2)
                throw new ArgumentNullException();
            for (int i = 0; i < participantsProposals.Count; i++)
                if (participantsProposals[i] == null || participantsProposals[i].Length == 0)
                    throw new ArgumentNullException();
            
            int k = participantsProposals.Count;

            double CRb = 0;
            double numTests = 0;
            for (int i = 0; i < k; i++)
                for (int j = i + 1; j < k; j++)
                    for (int p = 0; p < participantsProposals[i].Length; p++)
                        for (int q = 0; q < participantsProposals[j].Length; q++)
                        {
                            if (participantsProposals[i][p].CompareTo(participantsProposals[j][q]) == 0)
                                CRb++;
                            numTests++;
                        }
          
            CRb = CRb / numTests;
            return CRb;
        }

        #endregion

        #region confidence intervals for AR and DR (bootstrapping)

        /// <summary>
        /// Computes confidence intervals for agreement rate measures using 
        /// boostrapping and the percentile interval technique. 
        /// 95% CIs are computed by default (i.e., alpha = 0.05).
        /// </summary>
        private static double[] ComputeConfidenceInterval(string[] participantsProposals, AgreementMeasure measure, double alpha = 0.05)
        {
            int BOOTSTRAP_REPETITIONS = 5000;
            double[] distribution = new double[BOOTSTRAP_REPETITIONS];
            for (int indexRepetition = 0; indexRepetition < BOOTSTRAP_REPETITIONS; indexRepetition++)
            {
                string[] proposals = ResampleWithRepetition(participantsProposals);
                distribution[indexRepetition] = measure(proposals);
            }
            Array.Sort(distribution);
            
            return new double[]{
                distribution[(int)(distribution.Length * alpha / 2)],
                distribution[(int)(distribution.Length * (1.0 - alpha / 2))]
            };
        }

        private static Random random = new Random();
        private static string[] ResampleWithRepetition(string[] participantsProposals)
        {
            string[] proposals = new string[participantsProposals.Length];
            for (int i = 0; i < proposals.Length; i++)
                proposals[i] = participantsProposals[random.Next(participantsProposals.Length)];
            return proposals;
        }

        #endregion
    }
}
