﻿/**
 * AgreementRates.DLL (.NET Framework 4.5 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 * 	    MintViz Lab | MANSiD Research Center
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 * 	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publications for the Agreement Rates toolkit, and what should be 
 * used to cite it, are:
 *
 *  Radu-Daniel Vatavu, Jacob O. Wobbrock. (2016). 
 *  Between-subjects elicitation studies: Formalization and tool support.
 *  In Proceedings of CHI '16, the 34th ACM Conference on Human Factors in Computing Systems
 *  (San Jose, California, USA, May 2016). New York: ACM Press.
 *	
 *  Radu-Daniel Vatavu, Jacob O. Wobbrock. (2015). 
 *  Formalizing Agreement Analysis for Elicitation Studies: New Measures, Significance Test, and Toolkit.
 *  In Proceedings of CHI'15, the 33rd ACM SIGCHI Conference on Human Factors in Computing Systems 
 *  (Seoul, Republic of Korea, April 2015). New York: ACM Press.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2015-206, Radu-Daniel Vatavu and Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, nor 
 *	    University of Washington, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Jacob O. Wobbrock 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/

using System;
using System.Collections.Generic;
using System.Numerics;

namespace AgreementRates
{
    /// <summary>
    /// 
    ///  Implements statistical procedures to test the effect of referents and study groups on Agreement Rates 
    ///  for repeated-measures and between-subject experimental designs.
    ///  
    /// </summary>
    public class AgreementTestsOfSignificance
    {
        #region significance test for repeated-measures experimental designs (Vrd)

        #region chi-square critical values

        /// <summary>
        /// Stores the critical values of the chi^2 distribution 
        /// for a given number of degrees of freedom (df) at .05, .01, and .001 levels of significance.
        /// </summary>
        private class ChiSquareCriticalValue
        {
            public int df;
            public double p_05;
            public double p_01;
            public double p_001;

            public ChiSquareCriticalValue(int df, double p_05, double p_01, double p_001)
            {
                this.df = df;
                this.p_05 = p_05;
                this.p_01 = p_01;
                this.p_001 = p_001;
            }
        }

        private static ChiSquareCriticalValue[] chiSquare = new ChiSquareCriticalValue[]
        {
            new ChiSquareCriticalValue(0, 0, 0, 0),             // this entry is not used; it exists here only to index chiSquare correctly by df
        	new ChiSquareCriticalValue(1,3.84,6.64,10.83),
	        new ChiSquareCriticalValue(2,5.99,9.21,13.82),
	        new ChiSquareCriticalValue(3,7.82,11.35,16.27),
	        new ChiSquareCriticalValue(4,9.49,13.28,18.47),
	        new ChiSquareCriticalValue(5,11.07,15.09,20.52),
	        new ChiSquareCriticalValue(6,12.59,16.81,22.46),
	        new ChiSquareCriticalValue(7,14.07,18.48,24.32),
	        new ChiSquareCriticalValue(8,15.51,20.09,26.13),
	        new ChiSquareCriticalValue(9,16.92,21.67,27.88),
	        new ChiSquareCriticalValue(10,18.31,23.21,29.59),
	        new ChiSquareCriticalValue(11,19.68,24.73,31.26),
	        new ChiSquareCriticalValue(12,21.03,26.22,32.91),
	        new ChiSquareCriticalValue(13,22.36,27.69,34.53),
	        new ChiSquareCriticalValue(14,23.69,29.14,36.12),
	        new ChiSquareCriticalValue(15,25.00,30.58,37.70),
	        new ChiSquareCriticalValue(16,26.30,32.00,39.25),
	        new ChiSquareCriticalValue(17,27.59,33.41,40.79),
	        new ChiSquareCriticalValue(18,28.87,34.81,42.31),
	        new ChiSquareCriticalValue(19,30.14,36.19,43.82),
	        new ChiSquareCriticalValue(20,31.41,37.57,45.32),
	        new ChiSquareCriticalValue(21,32.67,38.93,46.80),
	        new ChiSquareCriticalValue(22,33.92,40.29,48.27),
	        new ChiSquareCriticalValue(23,35.17,41.64,49.73),
	        new ChiSquareCriticalValue(24,36.42,42.98,51.18),
	        new ChiSquareCriticalValue(25,37.65,44.31,52.62),
	        new ChiSquareCriticalValue(26,38.89,45.64,54.05),
	        new ChiSquareCriticalValue(27,40.11,46.96,55.48),
	        new ChiSquareCriticalValue(28,41.34,48.28,56.89),
	        new ChiSquareCriticalValue(29,42.56,49.59,58.30),
	        new ChiSquareCriticalValue(30,43.77,50.89,59.70),
	        new ChiSquareCriticalValue(31,44.99,52.19,61.10),
	        new ChiSquareCriticalValue(32,46.19,53.49,62.49),
	        new ChiSquareCriticalValue(33,47.40,54.78,63.87),
	        new ChiSquareCriticalValue(34,48.60,56.06,65.25),
	        new ChiSquareCriticalValue(35,49.80,57.34,66.62),
	        new ChiSquareCriticalValue(36,51.00,58.62,67.99),
	        new ChiSquareCriticalValue(37,52.19,59.89,69.35),
	        new ChiSquareCriticalValue(38,53.38,61.16,70.70),
	        new ChiSquareCriticalValue(39,54.57,62.43,72.06),
	        new ChiSquareCriticalValue(40,55.76,63.69,73.40),
	        new ChiSquareCriticalValue(41,56.94,64.95,74.75),
	        new ChiSquareCriticalValue(42,58.12,66.21,76.08),
	        new ChiSquareCriticalValue(43,59.30,67.46,77.42),
	        new ChiSquareCriticalValue(44,60.48,68.71,78.75),
	        new ChiSquareCriticalValue(45,61.66,69.96,80.08),
	        new ChiSquareCriticalValue(46,62.83,71.20,81.40),
	        new ChiSquareCriticalValue(47,64.00,72.44,82.72),
	        new ChiSquareCriticalValue(48,65.17,73.68,84.04),
	        new ChiSquareCriticalValue(49,66.34,74.92,85.35),
	        new ChiSquareCriticalValue(50,67.51,76.15,86.66)
        };

        #endregion

        /// <summary>
        /// Runs the V_{rd} statistical significance test for k referents and returns
        /// the value of the V_{rd} statistic and the level of significance (p).
        /// </summary>
        public static double Vrd(List<string[]> participantsProposals, out double p)
        {
            int k = participantsProposals.Count;  // number of referents

            // preliminary tests
            if (k < 2)
                throw new ArgumentException(String.Format("At least 2 referents are required to perform this test."));
            for (int i = 0; i < k; i++)
                if (participantsProposals[i] == null || participantsProposals[i].Length <= 1)
                    throw new ArgumentNullException();
            int P = participantsProposals[0].Length;   // number of participants (or proposals) from which options were elicited for each referent
            for (int i = 1; i < k; i++)
                if (participantsProposals[i].Length != P)
                    throw new ArgumentException(String.Format("Array lengths not equal: {0} vs. {1}", P, participantsProposals[i].Length));

            // compute agreement rates
            double[] ar = new double[k];
            for (int i = 0; i < k; i++)
                ar[i] = AgreementMeasures.AR(participantsProposals[i]);

            // compute the numerator
            double sumAR = 0, sumARSquared = 0;
            for (int i = 0; i < k; i++)
            {
                sumAR += ar[i];
                sumARSquared += ar[i] * ar[i];
            }
            double s1 = sumARSquared - 1.0 / k * sumAR * sumAR;

            // compute the denominator
            double sumCR = 0;
            for (int i = 0; i < k; i++)
                for (int j = 0; j < k; j++)
                    sumCR += AgreementMeasures.CR(participantsProposals[i], participantsProposals[j]);
            double s2 = sumAR - 1.0 / k * sumCR;

            // special case, check for denominator 0
            if (Math.Abs(s2) < 1e-3)
            {
                p = 1.0;
                return Double.NaN;
            }

            // compute the Vrd statistic
            double V = (k - 1) * (P * (P - 1) / 2.0) * s1 / s2;
            int df = k - 1;

            if (V > chiSquare[df].p_001)
                p = .001;
            else
                if (V > chiSquare[df].p_01)
                    p = .01;
                else
                    if (V > chiSquare[df].p_05)
                        p = .05;
                    else p = 1.00;
            return V;
        }

        /// <summary>
        /// Runs the V_{rd} significance test for k=2 referents and returns
        /// the value of the V_{rd} statistic and the level of significance (p).
        /// </summary>
        public static double Vrd(string[] participantsProposals_r1, string[] participantsProposals_r2, out double p)
        {
            if (participantsProposals_r1 == null || participantsProposals_r1.Length <= 1 ||
                participantsProposals_r2 == null || participantsProposals_r2.Length <= 1)
                throw new ArgumentNullException();
            if (participantsProposals_r1.Length != participantsProposals_r2.Length)
                throw new ArgumentException(String.Format("Array lengths not equal: {0} vs. {1}", participantsProposals_r1.Length, participantsProposals_r2.Length));

            int P = participantsProposals_r1.Length;
            double ar1 = AgreementMeasures.AR(participantsProposals_r1);
            double ar2 = AgreementMeasures.AR(participantsProposals_r2);
            double cr = AgreementMeasures.CR(participantsProposals_r1, participantsProposals_r2);

            // special case, check for denominator 0
            if (Math.Abs(ar1 + ar2 - 2 * cr) < 1e-3)
            {
                p = 1.0;
                return Double.NaN;
            }

            // compute the Vrd statistic
            double V = P * (P - 1) / 2.0 * (ar1 - ar2) * (ar1 - ar2) / (ar1 + ar2 - 2 * cr);
            int df = 1;
            if (V > chiSquare[df].p_001)
                p = .001;
            else
                if (V > chiSquare[df].p_01)
                    p = .01;
                else
                    if (V > chiSquare[df].p_05)
                        p = .05;
                    else p = 1.00;
            return V;
        }

        /// <summary>
        /// Runs the V_{rd} statistic to check for significance of an agreement rate against zero.
        /// Returns the value of the V_{rd} statistic and the level of significance (p).
        /// </summary>
        public static double Vrd(string[] participantsProposals, out double p)
        {
            if (participantsProposals == null || participantsProposals.Length <= 1)
                throw new ArgumentNullException();

            // compute the Vrd statistic
            int P = participantsProposals.Length;
            double V = P * (P - 1) / 2.0 * AgreementMeasures.AR(participantsProposals);
            int df = 1;
            if (V > chiSquare[df].p_001)
                p = .001;
            else
                if (V > chiSquare[df].p_01)
                    p = .01;
                else
                    if (V > chiSquare[df].p_05)
                        p = .05;
                    else p = 1.00;
            return V;
        }

        #endregion

        #region significance test for between-subjects experimental designs (Vb)

        /// <summary>
        /// Runs the Vb exact test of significance for agreement rates 
        /// with proposals elicited from k independent groups of participants.
        /// </summary>
        public static decimal Vb(List<string[]> proposalsPerGroup, out decimal probSignificance, bool useConditionalMargin = true)
        {
            // compute the agreement in each group
            int k = proposalsPerGroup.Count;  // number of groups
            int[] P = new int[k];             // number of participants in each group
            int[] a = new int[k];             // number of pairs in agreement in each group
            int A = 0;                        // total number of pairs in agreement (i.e., the marginal row total of the contingency table)
            for (int i = 0; i < k; i++)
            {
                P[i] = proposalsPerGroup[i].Length;
                a[i] = (int)(P[i] * (P[i] - 1) / 2 * AgreementMeasures.AR(proposalsPerGroup[i]));
                A += a[i];
            }
            
            // compute the frequencies of observing all possible agreement levels for each group size
            List<BigInteger[]> frequencyOfAgreement = new List<BigInteger[]>();
            for (int i = 0; i < k; i++)
                frequencyOfAgreement.Add(AgreementFrequencyDistribution.GenerateFrequencyTable(P[i]));

            // compute the deviation of the observed data from the null hypothesis
            // and its probability of occurence (not normalized at this point, but will be normalized at the end of this procedure)
            decimal devObserved = DeviationFromNullHypothesis(a, P);
            BigInteger pObserved = 1;
            for (int i = 0; i < k; i++)
                pObserved *= frequencyOfAgreement[i][a[i]];

            BigInteger pTotal = 0;
            BigInteger pTest = 0;

            // generate all possible contingency tables
            int[] stack = new int[k];   // implements the stack to generate all possible agreement levels for each group
            int l = 0;                  // current level of the stack
            stack[0] = -1;
            while (l >= 0)
            {
                bool canContinue = true;
                do
                {
                    stack[l]++;
                    if (stack[l] >= frequencyOfAgreement[l].Length)
                        canContinue = false;
                } while (canContinue && frequencyOfAgreement[l][stack[l]] == 0);

                if (canContinue)
                {
                    if (l == k - 1)
                    {
                        int sum = 0;
                        for (int i = 0; i <= l; i++)
                            sum += stack[i];
                        if (!useConditionalMargin || (useConditionalMargin && (sum == A)))
                        {
                            // compute the probability of this configuration and the deviation from the null hypothesis 
                            BigInteger p = 1;
                            for (int i = 0; i <= l; i++)
                                p *= frequencyOfAgreement[i][stack[i]];
                            decimal dev = DeviationFromNullHypothesis(stack, P);

                            pTotal += p;    // will be used to normalize probabilities in [0..1]
                            if (dev >= devObserved)
                                pTest += p; // add up all the probabilities for agreement configurations more extreme than our observed data
                        }
                    }
                    else
                    {
                        l++;
                        stack[l] = -1;
                    }
                }
                else
                    l--;
            }

            // implement the mid-P technique (because our observed data has also been generated 
            // with the stack among all the other configurations, subtract half of its probability here)
            probSignificance = Divide(pTest, pTotal) - Divide(pObserved, pTotal) / 2m; // normalize probability in [0..1]
            return devObserved;
        }

        private static decimal Divide(BigInteger a, BigInteger b)
        {
            int exponent = 5 + b.ToString().Length - a.ToString().Length + 1;
            a *= BigInteger.Pow(10, exponent);
            if (exponent > 20)
                return 0;
            return (decimal)(a / b) / (decimal)Math.Pow(10, exponent);
        }

        /// <summary>
        /// Computes deviation from the null hypothesis (i.e., how extreme the observed data is
        /// when considering the null hypothesis?)
        /// </summary>
        private static decimal DeviationFromNullHypothesis(int[] a, int[] P)
        {
            int[] n = new int[a.Length];
            int A = 0, N = 0;
            for (int i = 0; i < a.Length; i++)
            {
                A += a[i];
                n[i] = P[i] * (P[i] - 1) / 2;
                N += n[i];
            }

            decimal deviation = 0;
            for (int i = 0; i < a.Length; i++)
            {
                decimal expected = A * n[i] * 1m / N;
                deviation += (a[i] - expected) * 1m * (a[i] - expected);
            }
            return deviation;
        }

        #endregion
    }
}
