﻿/**
 * AgreementRates.DLL (.NET Framework 4.5 C# version)
 *
 * 	    Radu-Daniel Vatavu, Ph.D.
 * 	    MintViz Lab | MANSiD Research Center
 *	    University Stefan cel Mare of Suceava
 *	    Suceava 720229, Romania
 *	    vatavu@eed.usv.ro
 *
 * 	    Jacob O. Wobbrock, Ph.D.
 * 	    The Information School
 *	    University of Washington
 *	    Seattle, WA 98195-2840
 *	    wobbrock@uw.edu
 *
 * The academic publications for the Agreement Rates toolkit, and what should be 
 * used to cite it, are:
 *
 *  Radu-Daniel Vatavu, Jacob O. Wobbrock. (2016). 
 *  Between-subjects elicitation studies: Formalization and tool support.
 *  In Proceedings of CHI '16, the 34th ACM Conference on Human Factors in Computing Systems
 *  (San Jose, California, USA, May 2016). New York: ACM Press.
 *	
 *  Radu-Daniel Vatavu, Jacob O. Wobbrock. (2015). 
 *  Formalizing Agreement Analysis for Elicitation Studies: New Measures, Significance Test, and Toolkit.
 *  In Proceedings of CHI'15, the 33rd ACM SIGCHI Conference on Human Factors in Computing Systems 
 *  (Seoul, Republic of Korea, April 2015). New York: ACM Press.
 *
 * This software is distributed under the "New BSD License" agreement:
 *
 * Copyright (c) 2015-206, Radu-Daniel Vatavu and Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the names of the University Stefan cel Mare of Suceava, nor 
 *	    University of Washington, nor the names of its contributors 
 *	    may be used to endorse or promote products derived from this software 
 *	    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Radu-Daniel Vatavu OR Jacob O. Wobbrock 
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/

using System;
using System.Numerics;

namespace AgreementRates
{
    /// <summary>
    /// 
    /// Implements functions that return the frequency (probability) distributions for agreement rates.
    /// 
    /// </summary>
    class AgreementFrequencyDistribution
    {
        /// <summary>
        /// Returns the frequency distribution for all possible agreement levels that can be obtained with P participants.
        /// </summary>
        public static BigInteger[] GenerateFrequencyTable(int P)
        {
            BinomialCoefficients bc = new BinomialCoefficients(P);

            int n = P * (P - 1) / 2;                    // number of distinct pairs of participants
            BigInteger[] freq = new BigInteger[n + 1];  // stores the frequency of each agreement level that can be obtained with P participants
            Array.Clear(freq, 0, freq.Length);

            // generates all sets of integers that add to P
            int[] stack = new int[P];      // stack that generates sets of integers 
            int l = 0;                     // stack's current level
            stack[l] = 0;                  // first level starts at 0
            while (l >= 0)
            {
                stack[l]++;
                int sum = 0;
                for (int i = 0; i <= l; i++)
                    sum += stack[i];

                if (sum == P)  // we have a set of integers that add up to P
                {
                    // count the number of pairs of participants that are in agreement
                    int agreement = 0;
                    for (int i = l; i >= 0; i--)
                        agreement += stack[i] * (stack[i] - 1) / 2;

                    // count the number of distinct ways in which stack[i] participants (i=0..k) can be selected from P participants
                    // to compute the frequency of observing this particular partition of agreement
                    BigInteger frequency = 1;
                    BigInteger factorialOfEqualGroups = 1;    // e.g., the frequency of the partition 4,4,2,2,2,1 must be divided by 2! * 3! (corresponding to subgroups of size 4 and 2)
                    for (int i = l, N = P, M = 1; i >= 0; i--)
                    {
                        frequency *= (BigInteger)bc.Get(N, stack[i]);
                        N -= stack[i];
                        if (i < l && stack[i] != stack[i + 1])
                            M = 1;
                        factorialOfEqualGroups *= M++;
                    }
                    frequency /= factorialOfEqualGroups;
                    freq[agreement] += frequency;
                }
                else
                    if (sum < P)
                    {
                        l++;
                        stack[l] = stack[l - 1] - 1;
                    }
                    else l--;
            }

            return freq;
        }
    }
}
